import inspect
import typing
from typing import Final, final

from wirio._service_lookup._parameter_information import (
    ParameterInformation,
)
from wirio._service_lookup._typed_type import TypedType


@final
class ConstructorInformation:
    _type_: Final[TypedType]

    def __init__(self, type_: TypedType) -> None:
        self._type_ = type_

    def invoke(self, parameter_values: list[object]) -> object:
        return self._type_.invoke(parameter_values)

    def get_parameters(self) -> list[ParameterInformation]:
        init_method = self._type_.to_type().__init__
        init_signature = inspect.signature(init_method)
        init_type_hints = typing.get_type_hints(init_method, include_extras=True)  # pyright: ignore[reportUnusedVariable]
        parameter_informations: list[ParameterInformation] = []

        for parameter_name, parameter in init_signature.parameters.items():
            if parameter_name in ["self", "args", "kwargs"]:
                continue

            parameter_to_use = parameter.replace(
                annotation=init_type_hints[parameter_name]
            )
            parameter_information = ParameterInformation(parameter_to_use)
            parameter_informations.append(parameter_information)

        return parameter_informations



from typing import final

from wirio._service_lookup._typed_type import TypedType


class WirioError(Exception):
    """Base exception for Wirio Dependency Injection."""


@final
class ObjectDisposedError(WirioError):
    """The exception that is thrown when an operation is performed on a disposed object."""

    def __init__(self) -> None:
        super().__init__("BaseServiceProvider")


@final
class NonKeyedDescriptorMisuseError(WirioError):
    """The exception that is thrown when a service descriptor is not keyed."""

    def __init__(self) -> None:
        message = "This service descriptor is not keyed"
        super().__init__(message)


@final
class InvalidServiceDescriptorError(WirioError):
    """The exception that is thrown when a service descriptor is invalid."""

    def __init__(self) -> None:
        super().__init__("Invalid service descriptor")


@final
class InvalidServiceKeyTypeError(WirioError):
    """The exception that is thrown when the type of the key used for lookup doesn't match the type in the constructor parameter with the ServiceKey annotation metadata."""

    def __init__(self) -> None:
        super().__init__(
            "The type of the key used for lookup doesn't match the type in the constructor parameter with the ServiceKey annotation metadata"
        )


@final
class CannotResolveServiceError(WirioError):
    """The exception that is thrown when a service cannot be resolved."""

    def __init__(
        self, parameter_type: TypedType, implementation_type: TypedType
    ) -> None:
        message = f"Unable to resolve service for type '{parameter_type}' while attempting to activate '{implementation_type}'"
        super().__init__(message)


@final
class CannotResolveParameterServiceFromImplementationFactoryError(WirioError):
    """The exception that is thrown when a service for a parameter of an implementation factory cannot be resolved."""

    def __init__(self, parameter_type: TypedType) -> None:
        message = f"Unable to resolve service for type '{parameter_type}' while attempting to invoke an implementation factory"
        super().__init__(message)


@final
class CannotResolveServiceFromEndpointError(WirioError):
    """The exception that is thrown when a service for a parameter of an endpoint cannot be resolved."""

    def __init__(self, parameter_type: TypedType) -> None:
        message = f"Unable to resolve service for type '{parameter_type}' while attempting to invoke endpoint"
        super().__init__(message)


@final
class NoServiceRegisteredError(WirioError):
    """The exception that is thrown when no service is registered for a given type."""

    def __init__(self, service_type: TypedType) -> None:
        message = f"No service for type '{service_type}' has been registered"
        super().__init__(message)


@final
class NoKeyedServiceRegisteredError(WirioError):
    """The exception that is thrown when no keyed service is registered for a given type and key."""

    def __init__(self, service_type: TypedType, service_key_type: type) -> None:
        message = f"No keyed service for type '{service_type}' using key type '{service_key_type}' has been registered"
        super().__init__(message)


@final
class KeyedServiceAnyKeyUsedToResolveServiceError(WirioError):
    """The exception that is thrown when KeyedService.AnyKey is used to resolve a single service."""

    def __init__(self) -> None:
        message = "KeyedService.ANY_KEY cannot be used to resolve a single service"
        super().__init__(message)


@final
class CircularDependencyError(WirioError):
    """The exception that is thrown when a circular dependency is detected."""

    def __init__(self, service_type: TypedType) -> None:
        message = f"A circular dependency was detected for the service of type '{service_type}'"
        super().__init__(message)


@final
class NoSingletonServiceRegisteredError(WirioError):
    """The exception that is thrown when no singleton service is registered for a given type."""

    def __init__(self, service_type: TypedType) -> None:
        message = f"No singleton service for type '{service_type}' has been registered"
        super().__init__(message)


@final
class NoKeyedSingletonServiceRegisteredError(WirioError):
    """The exception that is thrown when no keyed singleton service is registered for a given type and key."""

    def __init__(self, service_type: TypedType, service_key_type: type) -> None:
        message = f"No keyed singleton service for type '{service_type}' using key type '{service_key_type}' has been registered"
        super().__init__(message)


@final
class ServiceProviderNotFullyInitializedError(WirioError):
    def __init__(self, method_name: str) -> None:
        message = f"The service provider must be fully initialized before calling '{method_name}'. Use 'async with' or '__aenter__'"
        super().__init__(message)



from typing import Final, final, override

from wirio._service_lookup._call_site_kind import CallSiteKind
from wirio._service_lookup._result_cache import ResultCache
from wirio._service_lookup._service_call_site import ServiceCallSite
from wirio._service_lookup._typed_type import TypedType


@final
class ConstantCallSite(ServiceCallSite):
    _service_type: Final[TypedType]

    def __init__(
        self,
        service_type: TypedType,
        default_value: object | None,
        service_key: object | None = None,
    ) -> None:
        result_cache = ResultCache.none(service_type=service_type)
        self._service_type = service_type

        if default_value is not None and not isinstance(
            default_value, service_type.to_type()
        ):
            error_message = f"Constant value of type '{type(default_value)}' can't be converted to service type '{service_type.to_type()}'"
            raise TypeError(error_message)

        super().__init__(cache=result_cache, key=service_key, value=default_value)

    @property
    @override
    def service_type(self) -> TypedType:
        return self._service_type

    @property
    @override
    def kind(self) -> CallSiteKind:
        return CallSiteKind.CONSTANT

    @property
    def default_value(self) -> object:
        return self._value



from abc import ABC, abstractmethod

from wirio.abstractions.service_provider_is_service import (
    ServiceProviderIsService,
)


class ServiceProviderIsKeyedService(ServiceProviderIsService, ABC):
    """Provide methods to determine if the specified service type with the specified service key is available from the :class:`BaseServiceProvider`."""

    @abstractmethod
    def is_keyed_service(self, service_key: object | None, service_type: type) -> bool:
        """Determine if the specified service type with the specified service key is available from the :class:`BaseServiceProvider`."""
        ...



import contextlib
import importlib
from dataclasses import dataclass
from typing import final, overload

from wirio.abstractions.service_key_lookup_mode import (
    ServiceKeyLookupMode,
)
from wirio.wirio_undefined import WirioUndefined


class Injectable:
    pass


@final
class FromServicesInjectable(Injectable):
    pass


@final
class ServiceKeyInjectable(Injectable):
    pass


@dataclass(frozen=True)
class FromKeyedServicesInjectable(Injectable):
    key: object | None
    lookup_mode: ServiceKeyLookupMode


def _return_injectable[TInjectable: Injectable](
    injectable: type[TInjectable],
) -> TInjectable:
    def _dependency() -> TInjectable:
        return injectable()

    _dependency.__is_wirio_depends__ = True  # pyright: ignore[reportFunctionMemberAccess]

    # In case of using FastAPI, wrap the injectable in a Depends
    with contextlib.suppress(ModuleNotFoundError):
        return importlib.import_module("fastapi").Depends(_dependency, use_cache=False)

    return _dependency()


def FromServices() -> Injectable:  # noqa: N802
    """Specify that a parameter should be bound using the requested service."""
    return _return_injectable(FromServicesInjectable)


def ServiceKey() -> ServiceKeyInjectable:  # noqa: N802
    """Specify the parameter to inject the key that was used for registration or resolution."""
    return _return_injectable(ServiceKeyInjectable)


@overload
def FromKeyedServices() -> FromKeyedServicesInjectable: ...


@overload
def FromKeyedServices(key: object | None) -> FromKeyedServicesInjectable: ...


def FromKeyedServices(  # noqa: N802
    key: object | None = WirioUndefined.INSTANCE,
) -> FromKeyedServicesInjectable:
    """Indicate that the parameter should be bound using the keyed service registered with the specified key."""

    def _dependency() -> FromKeyedServicesInjectable:
        is_key_not_provided = key is WirioUndefined.INSTANCE

        if is_key_not_provided:
            return FromKeyedServicesInjectable(
                key=None, lookup_mode=ServiceKeyLookupMode.INHERIT_KEY
            )

        lookup_mode = (
            ServiceKeyLookupMode.NULL_KEY
            if key is None
            else ServiceKeyLookupMode.EXPLICIT_KEY
        )
        return FromKeyedServicesInjectable(key=key, lookup_mode=lookup_mode)

    _dependency.__is_wirio_depends__ = True  # pyright: ignore[reportFunctionMemberAccess]

    # In case of using FastAPI, wrap the injectable in a Depends
    with contextlib.suppress(ModuleNotFoundError):
        return importlib.import_module("fastapi").Depends(_dependency, use_cache=False)

    return _dependency()



from types import TracebackType
from typing import Protocol, Self, runtime_checkable


@runtime_checkable
class SupportsAsyncContextManager(Protocol):
    async def __aenter__(self) -> Self: ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None: ...



from typing import Final, final, override

from wirio._service_lookup._call_site_kind import CallSiteKind
from wirio._service_lookup._result_cache import ResultCache
from wirio._service_lookup._service_call_site import ServiceCallSite
from wirio._service_lookup._typed_type import TypedType
from wirio.abstractions.base_service_provider import BaseServiceProvider


@final
class ServiceProviderCallSite(ServiceCallSite):
    _service_type: Final[TypedType]

    def __init__(self) -> None:
        service_type = TypedType.from_type(BaseServiceProvider)
        result_cache = ResultCache.none(service_type=service_type)
        self._service_type = service_type
        super().__init__(cache=result_cache, key=None)

    @property
    @override
    def service_type(self) -> TypedType:
        return self._service_type

    @property
    @override
    def kind(self) -> CallSiteKind:
        return CallSiteKind.SERVICE_PROVIDER



import typing
from inspect import Parameter
from types import UnionType
from typing import Annotated, Any, Final, Union, final

from wirio._service_lookup._typed_type import TypedType
from wirio._utils._param_utils import ParamUtils
from wirio.annotations import Injectable


@final
class ParameterInformation:
    _parameter_type: TypedType
    _is_optional: bool
    _has_default_value: bool
    _default_value: object | None
    _injectable_dependency: Final[Injectable | None]

    def __init__(self, parameter: Parameter) -> None:
        if parameter.annotation is Parameter.empty:
            error_message = (
                f"The parameter '{parameter.name}' must have a type annotation"
            )
            raise RuntimeError(error_message)

        self._is_optional = False
        self._has_default_value = False
        self._default_value = None
        self._injectable_dependency = None
        origin = typing.get_origin(parameter.annotation)
        args = typing.get_args(parameter.annotation)

        if origin is Annotated:
            parameter_type = args[0]
            self._has_default_value = True

            if callable(args[1]):
                self._default_value = args[1]()
            else:
                injectable_dependency = ParamUtils.get_injectable_dependency(parameter)

                if injectable_dependency is not None:
                    self._has_default_value = False
                    self._injectable_dependency = injectable_dependency
                else:
                    self._default_value = args[1]

            origin = typing.get_origin(parameter_type)

            if self._is_origin_a_union(origin):
                args = typing.get_args(parameter_type)
                self._extract_from_union(parameter, args)
            else:
                self._parameter_type = TypedType.from_type(parameter_type)
        elif self._is_origin_a_union(origin):
            self._extract_from_union(parameter, args)
        else:
            self._parameter_type = TypedType.from_type(parameter.annotation)

            if parameter.default is not Parameter.empty:
                self._has_default_value = True
                self._default_value = parameter.default

    @property
    def parameter_type(self) -> TypedType:
        return self._parameter_type

    @property
    def is_optional(self) -> bool:
        return self._is_optional

    @property
    def has_default_value(self) -> bool:
        return self._has_default_value

    @property
    def default_value(self) -> object | None:
        return self._default_value

    @property
    def injectable_dependency(self) -> Injectable | None:
        return self._injectable_dependency

    def _is_origin_a_union(self, origin: type) -> bool:
        return origin is Union or origin is UnionType

    def _extract_from_union(self, parameter: Parameter, args: tuple[Any, ...]) -> None:
        length_of_a_union_of_a_type_and_none = 2  # Example: int | None

        if len(args) > length_of_a_union_of_a_type_and_none:
            error_message = f"The parameter '{parameter.name}' has a Union type with more than one non-None type"
            raise RuntimeError(error_message)

        if len(args) > 1 and type(None) not in args:
            error_message = (
                f"The parameter '{parameter.name}' has a Union type without None"
            )
            raise RuntimeError(error_message)

        # At this point, we know that the Union has exactly one type plus None
        # The type that is not None is the actual parameter type
        self._parameter_type = TypedType.from_type(
            next(arg for arg in args if arg is not type(None))
        )

        self._is_optional = True

        if parameter.default is not Parameter.empty:
            self._has_default_value = True
            self._default_value = parameter.default



from typing import ClassVar, final


@final
class KeyedService:
    """Provide static APIs for use with `KeyedServiceProvider`."""

    @final
    class AnyKeyObj:
        def __str__(self) -> str:
            return "*"

    ANY_KEY: ClassVar[object] = AnyKeyObj()



import inspect
import typing
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from enum import Flag
from typing import (
    ClassVar,
    final,
    override,
)

from wirio._service_lookup._async_factory_call_site import (
    AsyncFactoryCallSite,
)
from wirio._service_lookup._call_site_visitor import CallSiteVisitor
from wirio._service_lookup._constant_call_site import (
    ConstantCallSite,
)
from wirio._service_lookup._constructor_call_site import (
    ConstructorCallSite,
)
from wirio._service_lookup._parameter_information import (
    ParameterInformation,
)
from wirio._service_lookup._service_call_site import (
    ServiceCallSite,
)
from wirio._service_lookup._service_provider_call_site import (
    ServiceProviderCallSite,
)
from wirio._service_lookup._supports_async_context_manager import (
    SupportsAsyncContextManager,
)
from wirio._service_lookup._supports_context_manager import (
    SupportsContextManager,
)
from wirio._service_lookup._sync_factory_call_site import (
    SyncFactoryCallSite,
)
from wirio._service_lookup._typed_type import TypedType
from wirio.annotations import FromKeyedServicesInjectable
from wirio.exceptions import (
    CannotResolveParameterServiceFromImplementationFactoryError,
)
from wirio.service_provider_engine_scope import (
    ServiceProviderEngineScope,
)
from wirio.wirio_undefined import WirioUndefined


class _RuntimeResolverLock(Flag):
    NONE = 0
    SCOPE = 1
    ROOT = 2


@dataclass(frozen=True)
class RuntimeResolverContext:
    scope: ServiceProviderEngineScope
    acquired_locks: _RuntimeResolverLock


@final
class CallSiteRuntimeResolver(CallSiteVisitor[RuntimeResolverContext, object | None]):
    INSTANCE: ClassVar["CallSiteRuntimeResolver"]

    async def resolve(
        self, call_site: ServiceCallSite, scope: ServiceProviderEngineScope
    ) -> object | None:
        # Fast path to avoid virtual calls if we already have the cached value in the root scope
        if scope.is_root_scope and call_site.value is not None:
            return call_site.value

        return await self._visit_call_site(
            call_site,
            RuntimeResolverContext(
                scope=scope, acquired_locks=_RuntimeResolverLock.NONE
            ),
        )

    @override
    async def _visit_root_cache(
        self, call_site: ServiceCallSite, argument: RuntimeResolverContext
    ) -> object | None:
        # If the value is already calculated, return it directly
        if call_site.value is not None:
            return call_site.value

        lock_type = _RuntimeResolverLock.ROOT
        service_provider_engine_scope = argument.scope.root_provider.root

        async with call_site.lock:
            # Lock the call_site and check if another coroutine already cached the value
            if call_site.value is not None:
                return call_site.value

            resolved_service = await self._visit_call_site_main(
                call_site=call_site,
                argument=RuntimeResolverContext(
                    scope=service_provider_engine_scope,
                    acquired_locks=argument.acquired_locks | lock_type,
                ),
            )
            await service_provider_engine_scope.capture_disposable(resolved_service)
            call_site.value = resolved_service
            return resolved_service

    @override
    async def _visit_scope_cache(
        self, call_site: ServiceCallSite, argument: RuntimeResolverContext
    ) -> object | None:
        # Check if we are in the situation where scoped service was promoted to singleton
        # and we need to lock the root
        if argument.scope.is_root_scope:
            return await self._visit_root_cache(call_site, argument)

        return await self._visit_cache(
            call_site, argument, argument.scope, _RuntimeResolverLock.SCOPE
        )

    async def _visit_cache(
        self,
        call_site: ServiceCallSite,
        argument: RuntimeResolverContext,
        service_provider_engine_scope: ServiceProviderEngineScope,
        lock_type: _RuntimeResolverLock,
    ) -> object | None:
        is_lock_taken = False
        resolved_services_lock = service_provider_engine_scope.resolved_services_lock
        resolved_services = service_provider_engine_scope.resolved_services

        # Taking locks only once allows us to fork resolution process
        # on another coroutine without causing the deadlock because we
        # always know that we are going to wait the other coroutine to finish before
        # releasing the lock
        if (argument.acquired_locks & lock_type) == _RuntimeResolverLock.NONE:
            await resolved_services_lock.acquire()
            is_lock_taken = True

        try:
            # Note: This method has already taken lock by the caller for resolution and access synchronization.
            # For scoped: takes a dictionary as both a resolution lock and a dictionary access lock.
            resolved_service = resolved_services.get(
                call_site.cache.key, WirioUndefined.INSTANCE
            )

            if resolved_service is not WirioUndefined.INSTANCE:
                return resolved_service

            resolved_service = await self._visit_call_site_main(
                call_site=call_site,
                argument=RuntimeResolverContext(
                    scope=service_provider_engine_scope,
                    acquired_locks=argument.acquired_locks | lock_type,
                ),
            )
            await service_provider_engine_scope.capture_disposable(resolved_service)
            resolved_services[call_site.cache.key] = resolved_service
            return resolved_service
        finally:
            if is_lock_taken:
                resolved_services_lock.release()

    @override
    async def _visit_dispose_cache(
        self, call_site: ServiceCallSite, argument: RuntimeResolverContext
    ) -> object | None:
        service = await self._visit_call_site_main(call_site, argument)
        return await argument.scope.capture_disposable(service)

    @override
    async def _visit_constructor(
        self,
        constructor_call_site: ConstructorCallSite,
        argument: RuntimeResolverContext,
    ) -> object:
        parameter_values: list[object | None] = []

        for parameter, parameter_call_site in zip(
            constructor_call_site.parameters,
            constructor_call_site.parameter_call_sites,
            strict=True,
        ):
            if parameter_call_site is None:
                if parameter.has_default_value:
                    parameter_values.append(parameter.default_value)
                    continue

                if parameter.is_optional:
                    parameter_values.append(None)
                    continue

                raise self._build_constructor_parameter_resolution_error(
                    parameter, constructor_call_site.service_type
                )

            parameter_service = await self._visit_call_site(
                parameter_call_site, argument
            )

            if parameter_service is None and not parameter.is_optional:
                raise self._build_constructor_parameter_resolution_error(
                    parameter, constructor_call_site.service_type
                )

            parameter_values.append(parameter_service)

        service = constructor_call_site.constructor_information.invoke(parameter_values)

        if isinstance(service, SupportsAsyncContextManager):
            await service.__aenter__()
        elif isinstance(service, SupportsContextManager):
            service.__enter__()

        return service

    @override
    def _visit_constant(
        self,
        constant_call_site: ConstantCallSite,
        argument: RuntimeResolverContext,
    ) -> object:
        return constant_call_site.default_value

    @override
    async def _visit_sync_factory(
        self,
        sync_factory_call_site: SyncFactoryCallSite,
        argument: RuntimeResolverContext,
    ) -> object | None:
        parameter_services = await self.get_parameter_services(
            sync_factory_call_site.implementation_factory, argument.scope
        )
        service = sync_factory_call_site.implementation_factory(*parameter_services)

        if isinstance(service, SupportsAsyncContextManager):
            await service.__aenter__()
        elif isinstance(service, SupportsContextManager):
            service.__enter__()

        return service

    @override
    async def _visit_async_factory(
        self,
        async_factory_call_site: AsyncFactoryCallSite,
        argument: RuntimeResolverContext,
    ) -> object | None:
        parameter_services = await self.get_parameter_services(
            async_factory_call_site.implementation_factory, argument.scope
        )
        service = await async_factory_call_site.implementation_factory(
            *parameter_services
        )

        if isinstance(service, SupportsAsyncContextManager):
            await service.__aenter__()
        elif isinstance(service, SupportsContextManager):
            service.__enter__()

        return service

    @override
    def _visit_service_provider(
        self,
        service_provider_call_site: ServiceProviderCallSite,
        argument: RuntimeResolverContext,
    ) -> object | None:
        return argument.scope

    async def get_parameter_services(
        self,
        implementation_factory: Callable[..., Awaitable[object]]
        | Callable[..., object],
        scope: ServiceProviderEngineScope,
    ) -> list[object | None]:
        parameter_services: list[object | None] = []
        signature = inspect.signature(implementation_factory)

        for parameter in signature.parameters.values():
            parameter_information = ParameterInformation(parameter)
            parameter_service: object | None = None

            if isinstance(
                parameter_information.injectable_dependency, FromKeyedServicesInjectable
            ):
                parameter_service = await scope.get_keyed_service_object(
                    parameter_information.injectable_dependency.key,
                    parameter_information.parameter_type,
                )
            else:
                parameter_service = await scope.get_service_object(
                    parameter_information.parameter_type
                )

            if parameter_service is None:
                if parameter_information.has_default_value:
                    parameter_services.append(parameter_information.default_value)
                elif parameter_information.is_optional:
                    parameter_services.append(None)
                else:
                    raise CannotResolveParameterServiceFromImplementationFactoryError(
                        parameter_information.parameter_type
                    )
            else:
                parameter_services.append(parameter_service)

        return parameter_services

    def _get_parameter_types(
        self,
        implementation_factory: Callable[..., Awaitable[object]]
        | Callable[..., object],
    ) -> list[type]:
        type_hints: dict[str, type] = typing.get_type_hints(implementation_factory)
        return [
            parameter_type
            for parameter_name, parameter_type in type_hints.items()
            if parameter_name != "return"
        ]

    def _build_constructor_parameter_resolution_error(
        self, parameter_information: ParameterInformation, service_type: TypedType
    ) -> RuntimeError:
        return RuntimeError(
            f"Unable to resolve service with type '{parameter_information.parameter_type}' while attempting to activate '{service_type}'"
        )


CallSiteRuntimeResolver.INSTANCE = CallSiteRuntimeResolver()



from typing import Final, final

from wirio._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from wirio._service_lookup._typed_type import TypedType
from wirio._service_lookup.call_site_result_cache_location import (
    CallSiteResultCacheLocation,
)
from wirio._service_lookup.service_cache_key import ServiceCacheKey
from wirio.service_lifetime import ServiceLifetime


@final
class ResultCache:
    """Track cached service."""

    _location: Final[CallSiteResultCacheLocation]
    _key: Final[ServiceCacheKey]

    def __init__(
        self, location: CallSiteResultCacheLocation, key: ServiceCacheKey
    ) -> None:
        self._location = location
        self._key = key

    @classmethod
    def from_lifetime(
        cls,
        lifetime: ServiceLifetime,
        service_identifier: ServiceIdentifier,
        slot: int,
    ) -> "ResultCache":
        match lifetime:
            case ServiceLifetime.SINGLETON:
                location = CallSiteResultCacheLocation.ROOT
            case ServiceLifetime.SCOPED:
                location = CallSiteResultCacheLocation.SCOPE
            case ServiceLifetime.TRANSIENT:
                location = CallSiteResultCacheLocation.DISPOSE

        key = ServiceCacheKey(service_identifier, slot)
        return cls(location, key)

    @classmethod
    def none(cls, service_type: TypedType) -> "ResultCache":
        cache_key = ServiceCacheKey(
            service_identifier=ServiceIdentifier.from_service_type(service_type), slot=0
        )
        return cls(CallSiteResultCacheLocation.NONE, cache_key)

    @property
    def location(self) -> CallSiteResultCacheLocation:
        return self._location

    @property
    def key(self) -> ServiceCacheKey:
        return self._key



import functools
import inspect
from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from contextvars import ContextVar
from inspect import Parameter
from typing import TYPE_CHECKING, Any, final

from fastapi import FastAPI
from fastapi.routing import APIRoute
from starlette.requests import Request
from starlette.routing import BaseRoute, Match
from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.websockets import WebSocket

from wirio._service_lookup._parameter_information import (
    ParameterInformation,
)
from wirio._utils._param_utils import ParamUtils
from wirio.abstractions.service_scope import ServiceScope
from wirio.annotations import FromKeyedServicesInjectable
from wirio.exceptions import CannotResolveServiceFromEndpointError

if TYPE_CHECKING:
    from wirio.service_collection import ServiceCollection
    from wirio.service_provider import ServiceProvider


_current_request: ContextVar[Request | WebSocket] = ContextVar(
    "wirio_starlette_request"
)


@final
class FastApiDependencyInjection:
    @classmethod
    def setup(cls, app: FastAPI, services: "ServiceCollection") -> None:
        app.state.wirio_services = services
        app.add_middleware(_WirioAsgiMiddleware)
        cls._update_lifespan(app)
        cls._inject_routes(app.routes)

    @classmethod
    def _update_lifespan(cls, app: FastAPI) -> None:
        old_lifespan = app.router.lifespan_context

        @asynccontextmanager
        async def new_lifespan(app: FastAPI) -> AsyncGenerator[Any]:
            services: ServiceCollection = app.state.wirio_services

            async with services.build_service_provider() as service_provider:
                app.state.wirio_service_provider = service_provider

                async with old_lifespan(app) as state:
                    yield state

        app.router.lifespan_context = new_lifespan

    @classmethod
    def _are_annotated_parameters_with_wirio_dependencies(
        cls, target: Callable[..., Any]
    ) -> bool:
        for parameter in inspect.signature(target).parameters.values():
            if (
                parameter.annotation is not None
                and hasattr(parameter.annotation, "__metadata__")
                and hasattr(parameter.annotation.__metadata__[0], "dependency")
                and hasattr(
                    parameter.annotation.__metadata__[0].dependency,
                    "__is_wirio_depends__",
                )
            ):
                return True

        return False

    @classmethod
    def _inject_routes(cls, routes: list[BaseRoute]) -> None:
        for route in routes:
            if not (
                isinstance(route, APIRoute)
                and route.dependant.call is not None
                and inspect.iscoroutinefunction(route.dependant.call)
                and cls._are_annotated_parameters_with_wirio_dependencies(
                    route.dependant.call
                )
            ):
                continue

            route.dependant.call = cls._inject_from_container(route.dependant.call)

    @classmethod
    def _inject_from_container(cls, target: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(target)
        async def _inject_async_target(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
            parameters_to_inject = cls._get_parameters_to_inject(target)
            parameters_to_inject_resolved: dict[str, Any] = {
                injected_parameter_name: await cls._resolve_injected_parameter(
                    parameter_information
                )
                for injected_parameter_name, parameter_information in parameters_to_inject.items()
            }
            return await target(*args, **{**kwargs, **parameters_to_inject_resolved})

        return _inject_async_target

    @classmethod
    def _get_request_container(cls) -> ServiceScope:
        """When inside a request, return the scoped container instance handling the current request.

        This is what we almost always want. It has all the information the app container has in addition
        to data specific to the current request.
        """
        return _current_request.get().state.wirio_service_scope

    @classmethod
    def _get_parameters_to_inject(
        cls, target: Callable[..., Any]
    ) -> dict[str, ParameterInformation]:
        result: dict[str, ParameterInformation] = {}

        for parameter_name, parameter in inspect.signature(target).parameters.items():
            if parameter.annotation is Parameter.empty:
                continue

            injectable_dependency = ParamUtils.get_injectable_dependency(parameter)

            if injectable_dependency is None:
                continue

            parameter_information = ParameterInformation(parameter=parameter)
            result[parameter_name] = parameter_information

        return result

    @classmethod
    async def _resolve_injected_parameter(
        cls, parameter_information: ParameterInformation
    ) -> object | None:
        parameter_service: object | None

        if isinstance(
            parameter_information.injectable_dependency, FromKeyedServicesInjectable
        ):
            parameter_service = await cls._get_request_container().service_provider.get_keyed_service_object(
                parameter_information.injectable_dependency.key,
                parameter_information.parameter_type,
            )
        else:
            parameter_service = (
                await cls._get_request_container().service_provider.get_service_object(
                    parameter_information.parameter_type
                )
            )

        if parameter_service is None:
            if parameter_information.is_optional:
                return None

            raise CannotResolveServiceFromEndpointError(
                parameter_information.parameter_type
            )

        return parameter_service


@final
class _WirioAsgiMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in {"http", "websocket"}:
            return await self.app(scope, receive, send)

        if scope["type"] == "http":
            request = Request(scope, receive=receive, send=send)
        else:
            request = WebSocket(scope, receive, send)

        token = _current_request.set(request)

        try:
            is_endpoint_matched = False
            is_async_endpoint = False

            for route in scope["app"].routes:
                if (
                    isinstance(route, APIRoute)
                    and route.matches(scope)[0] == Match.FULL
                ):
                    is_endpoint_matched = True
                    original = inspect.unwrap(route.dependant.call)  # pyright: ignore[reportArgumentType]
                    is_async_endpoint = inspect.iscoroutinefunction(original)
                    break

            if not is_endpoint_matched or not is_async_endpoint:
                return await self.app(scope, receive, send)

            services: ServiceProvider = request.app.state.wirio_service_provider

            async with services.create_scope() as service_scope:
                request.state.wirio_service_scope = service_scope
                await self.app(scope, receive, send)
        finally:
            _current_request.reset(token)



from typing import ClassVar, final


@final
class WirioUndefined:
    """A type used as a sentinel for undefined values."""

    INSTANCE: ClassVar["WirioUndefined"]


WirioUndefined.INSTANCE = WirioUndefined()



from abc import ABC, abstractmethod


class ServiceProviderIsService(ABC):
    """Provide methods to determine if the specified service type is available from the :class:`BaseServiceProvider`."""

    @abstractmethod
    def is_service(self, service_type: type) -> bool:
        """Determine if the specified service type is available from the :class:`BaseServiceProvider`."""
        ...



from collections.abc import Sequence
from inspect import Parameter
from typing import Any

from wirio.annotations import Injectable


class ParamUtils:
    @classmethod
    def get_injectable_dependency(cls, parameter: Parameter) -> Injectable | None:
        if not hasattr(parameter.annotation, "__metadata__"):
            return None

        metadata = cls._get_metadata(parameter)

        if metadata is None:
            return None

        metadata_item = metadata[0]

        if isinstance(metadata_item, Injectable):
            return metadata_item

        if hasattr(metadata_item, "dependency") and hasattr(
            metadata_item.dependency, "__is_wirio_depends__"
        ):
            dependency = metadata_item.dependency()  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType, reportAttributeAccessIssue]

            if isinstance(dependency, Injectable):
                return dependency

        return None

    @classmethod
    def _get_metadata(cls, parameter: Parameter) -> Sequence[Any] | None:
        if not hasattr(parameter.annotation, "__metadata__"):
            return None

        return parameter.annotation.__metadata__



import asyncio
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from typing import ClassVar, Final, final, override

from wirio._service_lookup._async_concurrent_dictionary import (
    AsyncConcurrentDictionary,
)
from wirio._service_lookup._async_factory_call_site import (
    AsyncFactoryCallSite,
)
from wirio._service_lookup._call_site_chain import CallSiteChain
from wirio._service_lookup._constant_call_site import (
    ConstantCallSite,
)
from wirio._service_lookup._constructor_call_site import (
    ConstructorCallSite,
)
from wirio._service_lookup._constructor_information import (
    ConstructorInformation,
)
from wirio._service_lookup._parameter_information import (
    ParameterInformation,
)
from wirio._service_lookup._result_cache import ResultCache
from wirio._service_lookup._service_call_site import (
    ServiceCallSite,
)
from wirio._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from wirio._service_lookup._sync_factory_call_site import (
    SyncFactoryCallSite,
)
from wirio._service_lookup._typed_type import TypedType
from wirio._service_lookup.service_cache_key import ServiceCacheKey
from wirio.abstractions.base_service_provider import BaseServiceProvider
from wirio.abstractions.keyed_service import KeyedService
from wirio.abstractions.service_key_lookup_mode import (
    ServiceKeyLookupMode,
)
from wirio.abstractions.service_provider_is_keyed_service import (
    ServiceProviderIsKeyedService,
)
from wirio.abstractions.service_provider_is_service import (
    ServiceProviderIsService,
)
from wirio.abstractions.service_scope_factory import (
    ServiceScopeFactory,
)
from wirio.annotations import (
    FromKeyedServicesInjectable,
    ServiceKeyInjectable,
)
from wirio.exceptions import (
    CannotResolveServiceError,
    InvalidServiceDescriptorError,
    InvalidServiceKeyTypeError,
)
from wirio.service_descriptor import ServiceDescriptor


@final
class _ServiceDescriptorCacheItem:
    _item: ServiceDescriptor | None
    _items: list[ServiceDescriptor] | None

    def __init__(self) -> None:
        self._item = None
        self._items = None

    @property
    def last(self) -> ServiceDescriptor:
        if self._items is not None and len(self._items) > 0:
            return self._items[len(self._items) - 1]

        assert self._item is not None
        return self._item

    def add(self, descriptor: ServiceDescriptor) -> "_ServiceDescriptorCacheItem":
        new_cache_item = _ServiceDescriptorCacheItem()

        if self._item is None:
            new_cache_item._item = descriptor
        else:
            new_cache_item._item = self._item
            new_cache_item._items = self._items if self._items is not None else []
            new_cache_item._items.append(descriptor)

        return new_cache_item


@dataclass(frozen=True)
class _ServiceOverride:
    exists: bool
    value: object | None = None


@final
class CallSiteFactory(ServiceProviderIsKeyedService, ServiceProviderIsService):
    _DEFAULT_SLOT: ClassVar[int] = 0

    _descriptors: Final[list[ServiceDescriptor]]
    _descriptor_lookup: Final[dict[ServiceIdentifier, _ServiceDescriptorCacheItem]]
    _call_site_cache: Final[AsyncConcurrentDictionary[ServiceCacheKey, ServiceCallSite]]
    _call_site_locks: Final[AsyncConcurrentDictionary[ServiceIdentifier, asyncio.Lock]]
    _service_overrides: Final[dict[ServiceIdentifier, list[object | None]]]

    def __init__(self, descriptors: list["ServiceDescriptor"]) -> None:
        self._descriptors = descriptors.copy()
        self._descriptor_lookup = {}
        self._call_site_cache = AsyncConcurrentDictionary[
            ServiceCacheKey, ServiceCallSite
        ]()
        self._call_site_locks = AsyncConcurrentDictionary[
            ServiceIdentifier, asyncio.Lock
        ]()
        self._service_overrides = {}
        self._populate()

    @override
    def is_service(self, service_type: type) -> bool:
        return self._is_service(
            ServiceIdentifier.from_service_type(
                service_type=TypedType.from_type(service_type)
            )
        )

    @override
    def is_keyed_service(self, service_key: object | None, service_type: type) -> bool:
        return self._is_service(
            ServiceIdentifier.from_service_type(
                service_type=TypedType.from_type(service_type), service_key=service_key
            )
        )

    async def get_call_site(
        self, service_identifier: ServiceIdentifier, call_site_chain: CallSiteChain
    ) -> ServiceCallSite | None:
        overridden_call_site = self._get_overridden_call_site(service_identifier)

        if overridden_call_site is not None:
            return overridden_call_site

        service_cache_key = ServiceCacheKey(service_identifier, self._DEFAULT_SLOT)
        service_call_site = self._call_site_cache.get(service_cache_key)

        if service_call_site is None:
            return await self._create_call_site(
                service_identifier=service_identifier, call_site_chain=call_site_chain
            )

        return service_call_site

    async def add(
        self, service_identifier: ServiceIdentifier, service_call_site: ServiceCallSite
    ) -> None:
        cache_key = ServiceCacheKey(service_identifier, self._DEFAULT_SLOT)
        await self._call_site_cache.upsert(key=cache_key, value=service_call_site)

    @contextmanager
    def override_service(
        self,
        service_identifier: ServiceIdentifier,
        implementation_instance: object | None,
    ) -> Generator[None]:
        self._add_override(service_identifier, implementation_instance)

        try:
            yield
        finally:
            self._remove_override(service_identifier)

    def get_overridden_call_site(
        self, service_identifier: ServiceIdentifier
    ) -> ServiceCallSite | None:
        return self._get_overridden_call_site(service_identifier)

    async def _create_call_site(
        self, service_identifier: ServiceIdentifier, call_site_chain: CallSiteChain
    ) -> ServiceCallSite | None:
        async def _create_new_lock(_: ServiceIdentifier) -> asyncio.Lock:
            return asyncio.Lock()

        # We need to lock the resolution process for a single service type at a time.
        # Consider the following:
        # C -> D -> A
        # E -> D -> A
        # Resolving C and E in parallel means that they will be modifying the callsite cache concurrently
        # to add the entry for C and E, but the resolution of D and A is synchronized
        # to make sure C and E both reference the same instance of the callsite.
        #
        # This is to make sure we can safely store singleton values on the callsites themselves

        call_site_lock = await self._call_site_locks.get_or_add(
            service_identifier, _create_new_lock
        )

        # Check if the lock is already acquired to prevent deadlocks in case of re-entrancy
        if call_site_lock.locked():
            call_site_chain.check_circular_dependency(service_identifier)

        async with call_site_lock:
            return await self._try_create_exact_from_service_identifier(
                service_identifier, call_site_chain
            )

    def _populate(self) -> None:
        for descriptor in self._descriptors:
            cache_key = ServiceIdentifier.from_descriptor(descriptor)
            cache_item = self._descriptor_lookup.get(
                cache_key, _ServiceDescriptorCacheItem()
            )
            self._descriptor_lookup[cache_key] = cache_item.add(descriptor)

    async def _try_create_exact_from_service_identifier(
        self, service_identifier: ServiceIdentifier, call_site_chain: CallSiteChain
    ) -> ServiceCallSite | None:
        service_descriptor_cache_item = self._descriptor_lookup.get(
            service_identifier, None
        )

        if service_descriptor_cache_item is not None:
            return await self._try_create_exact_from_service_descriptor(
                service_descriptor_cache_item.last,
                service_identifier,
                call_site_chain,
                self._DEFAULT_SLOT,
            )

        # Check if there is a registration with `KeyedService.ANY_KEY`
        if service_identifier.service_key is not None:
            catch_all_identifier = ServiceIdentifier(
                service_type=service_identifier.service_type,
                service_key=KeyedService.ANY_KEY,
            )

            service_descriptor_cache_item = self._descriptor_lookup.get(
                catch_all_identifier, None
            )

            if service_descriptor_cache_item is not None:
                return await self._try_create_exact_from_service_descriptor(
                    service_descriptor_cache_item.last,
                    service_identifier,
                    call_site_chain,
                    self._DEFAULT_SLOT,
                )

        return None

    async def _try_create_exact_from_service_descriptor(
        self,
        service_descriptor: ServiceDescriptor,
        service_identifier: ServiceIdentifier,
        call_site_chain: CallSiteChain,
        slot: int,
    ) -> ServiceCallSite | None:
        if not self._should_create_exact(
            service_descriptor.service_type, service_identifier.service_type
        ):
            return None

        return await self._create_exact(
            service_descriptor, service_identifier, call_site_chain, slot
        )

    def _should_create_exact(
        self, descriptor_type: TypedType, service_type: TypedType
    ) -> bool:
        return descriptor_type == service_type

    async def _create_exact(
        self,
        service_descriptor: ServiceDescriptor,
        service_identifier: ServiceIdentifier,
        call_site_chain: CallSiteChain,
        slot: int,
    ) -> ServiceCallSite:
        call_site_key = ServiceCacheKey(service_identifier, slot)
        service_call_site = self._call_site_cache.get(call_site_key)

        if service_call_site is not None:
            return service_call_site

        cache = ResultCache.from_lifetime(
            service_descriptor.lifetime, service_identifier, slot
        )

        if service_descriptor.has_implementation_instance():
            service_call_site = ConstantCallSite(
                service_type=service_descriptor.service_type,
                default_value=service_descriptor.get_implementation_instance(),
                service_key=service_descriptor.service_key,
            )
        elif (
            not service_descriptor.is_keyed_service
            and service_descriptor.sync_implementation_factory is not None
        ):
            service_call_site = SyncFactoryCallSite.from_implementation_factory(
                cache=cache,
                service_type=service_descriptor.service_type,
                implementation_factory=service_descriptor.sync_implementation_factory,
            )
        elif (
            service_descriptor.is_keyed_service
            and service_descriptor.keyed_sync_implementation_factory is not None
        ):
            service_call_site = SyncFactoryCallSite.from_keyed_implementation_factory(
                cache=cache,
                service_type=service_descriptor.service_type,
                implementation_factory=service_descriptor.keyed_sync_implementation_factory,
                service_key=service_identifier.service_key,
            )
        elif (
            not service_descriptor.is_keyed_service
            and service_descriptor.async_implementation_factory is not None
        ):
            service_call_site = AsyncFactoryCallSite.from_implementation_factory(
                cache=cache,
                service_type=service_descriptor.service_type,
                implementation_factory=service_descriptor.async_implementation_factory,
            )
        elif (
            service_descriptor.is_keyed_service
            and service_descriptor.keyed_async_implementation_factory is not None
        ):
            service_call_site = AsyncFactoryCallSite.from_keyed_implementation_factory(
                cache=cache,
                service_type=service_descriptor.service_type,
                implementation_factory=service_descriptor.keyed_async_implementation_factory,
                service_key=service_identifier.service_key,
            )
        elif service_descriptor.has_implementation_type():
            implementation_type = service_descriptor.get_implementation_type()
            assert implementation_type is not None
            service_call_site = await self._create_constructor_call_site(
                cache=cache,
                service_identifier=service_identifier,
                implementation_type=implementation_type,
                call_site_chain=call_site_chain,
            )
        else:
            raise InvalidServiceDescriptorError

        await self._call_site_cache.upsert(key=call_site_key, value=service_call_site)
        return service_call_site

    def _get_overridden_call_site(
        self, service_identifier: ServiceIdentifier
    ) -> ServiceCallSite | None:
        override = self._get_overridden_instance(service_identifier)

        if not override.exists:
            return None

        return ConstantCallSite(
            service_type=service_identifier.service_type,
            default_value=override.value,
            service_key=service_identifier.service_key,
        )

    def _get_overridden_instance(
        self, service_identifier: ServiceIdentifier
    ) -> _ServiceOverride:
        overrides = self._service_overrides.get(service_identifier)

        if overrides is not None and len(overrides) > 0:
            return _ServiceOverride(exists=True, value=overrides[-1])

        catch_all_identifier = self._get_catch_all_service_identifier(
            service_identifier
        )

        if catch_all_identifier is None:
            return _ServiceOverride(exists=False, value=None)

        overrides = self._service_overrides.get(catch_all_identifier)

        if overrides is not None and len(overrides) > 0:
            return _ServiceOverride(exists=True, value=overrides[-1])

        return _ServiceOverride(exists=False, value=None)

    def _get_catch_all_service_identifier(
        self, service_identifier: ServiceIdentifier
    ) -> ServiceIdentifier | None:
        if (
            service_identifier.service_key is None
            or service_identifier.service_key == KeyedService.ANY_KEY
        ):
            return None

        return ServiceIdentifier.from_service_type(
            service_type=service_identifier.service_type,
            service_key=KeyedService.ANY_KEY,
        )

    def _add_override(
        self,
        service_identifier: ServiceIdentifier,
        implementation_instance: object | None,
    ) -> None:
        overrides = self._service_overrides.setdefault(service_identifier, [])
        overrides.append(implementation_instance)

    def _remove_override(self, service_identifier: ServiceIdentifier) -> None:
        overrides = self._service_overrides.get(service_identifier)
        assert overrides is not None
        overrides.pop()

        if len(overrides) == 0:
            self._service_overrides.pop(service_identifier)

    async def _create_constructor_call_site(
        self,
        cache: ResultCache,
        service_identifier: ServiceIdentifier,
        implementation_type: TypedType,
        call_site_chain: CallSiteChain,
    ) -> ServiceCallSite:
        try:
            call_site_chain.add(service_identifier, implementation_type)
            parameter_call_sites: list[ServiceCallSite | None] | None = None
            constructor_information = ConstructorInformation(implementation_type)
            parameters = constructor_information.get_parameters()
            parameter_call_sites = await self._create_argument_call_sites(
                service_identifier=service_identifier,
                implementation_type=implementation_type,
                parameters=parameters,
                call_site_chain=call_site_chain,
            )
            return ConstructorCallSite(
                cache=cache,
                service_type=service_identifier.service_type,
                constructor_information=constructor_information,
                parameters=parameters,
                parameter_call_sites=parameter_call_sites,
                service_key=service_identifier.service_key,
            )
        finally:
            call_site_chain.remove(service_identifier)

    async def _create_argument_call_sites(  # noqa: C901, PLR0912
        self,
        service_identifier: ServiceIdentifier,
        implementation_type: TypedType,
        parameters: list[ParameterInformation],
        call_site_chain: CallSiteChain,
    ) -> list[ServiceCallSite | None]:
        if len(parameters) == 0:
            return []

        parameter_call_sites: list[ServiceCallSite | None] = []

        for parameter in parameters:
            call_site: ServiceCallSite | None = None
            is_keyed_parameter = False
            parameter_type = parameter.parameter_type

            if parameter.injectable_dependency is not None:
                if service_identifier.service_key is not None and isinstance(
                    parameter.injectable_dependency, ServiceKeyInjectable
                ):
                    # Even though the parameter may be strongly typed, support `object` if `ANY_KEY` is used

                    if service_identifier.service_key == KeyedService.ANY_KEY:
                        parameter_type = TypedType.from_type(object)
                    elif parameter_type.to_type() is not type(
                        service_identifier.service_key
                    ) and parameter_type.to_type() is not type(object):
                        raise InvalidServiceKeyTypeError

                    call_site = ConstantCallSite(
                        service_type=parameter_type,
                        default_value=service_identifier.service_key,
                    )
                elif isinstance(
                    parameter.injectable_dependency, FromKeyedServicesInjectable
                ):
                    service_key: object | None = None

                    match parameter.injectable_dependency.lookup_mode:
                        case ServiceKeyLookupMode.INHERIT_KEY:
                            service_key = service_identifier.service_key
                        case ServiceKeyLookupMode.EXPLICIT_KEY:
                            service_key = parameter.injectable_dependency.key
                        case ServiceKeyLookupMode.NULL_KEY:
                            service_key = None

                    if service_key is not None:
                        call_site = await self.get_call_site(
                            ServiceIdentifier.from_service_type(
                                service_type=parameter_type, service_key=service_key
                            ),
                            call_site_chain=call_site_chain,
                        )
                        is_keyed_parameter = True

            if not is_keyed_parameter and call_site is None:
                call_site = await self.get_call_site(
                    ServiceIdentifier.from_service_type(parameter_type), call_site_chain
                )

            if call_site is None and parameter.has_default_value:
                if parameter.is_optional:
                    parameter_call_sites.append(None)
                    continue

                call_site = ConstantCallSite(
                    service_type=parameter_type,
                    default_value=parameter.default_value,
                )

            if call_site is None and parameter.is_optional:
                parameter_call_sites.append(None)
                continue

            if call_site is None:
                raise CannotResolveServiceError(
                    parameter_type=parameter_type,
                    implementation_type=implementation_type,
                )

            parameter_call_sites.append(call_site)

        return parameter_call_sites

    def _is_service(self, service_identifier: ServiceIdentifier) -> bool:
        service_type = service_identifier.service_type

        if service_identifier in self._descriptor_lookup:
            return True

        if (
            service_identifier.service_key is not None
            and ServiceIdentifier(
                service_type=service_type, service_key=KeyedService.ANY_KEY
            )
            in self._descriptor_lookup
        ):
            return True

        return (
            service_type == TypedType.from_type(BaseServiceProvider)
            or service_type == TypedType.from_type(ServiceScopeFactory)
            or service_type == TypedType.from_type(ServiceProviderIsService)
            or service_type == TypedType.from_type(ServiceProviderIsKeyedService)
        )



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from wirio.abstractions.service_scope import ServiceScope


class ServiceScopeFactory(ABC):
    """Create instances of :class:`ServiceScope`, which is used to create services within a scope."""

    @abstractmethod
    def create_scope(self) -> "ServiceScope":
        """Create a :class:`ServiceScope` that contains a :class:`ServiceProvider` used to resolve dependencies from a newly created scope."""
        ...



import asyncio
from collections.abc import Awaitable, Callable, Generator
from contextlib import contextmanager
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Final, Self, final, override

from wirio._service_lookup._async_concurrent_dictionary import (
    AsyncConcurrentDictionary,
)
from wirio._service_lookup._call_site_chain import CallSiteChain
from wirio._service_lookup._call_site_factory import CallSiteFactory
from wirio._service_lookup._constant_call_site import (
    ConstantCallSite,
)
from wirio._service_lookup._runtime_service_provider_engine import (
    RuntimeServiceProviderEngine,
)
from wirio._service_lookup._service_call_site import (
    ServiceCallSite,
)
from wirio._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from wirio._service_lookup._service_provider_call_site import (
    ServiceProviderCallSite,
)
from wirio._service_lookup._service_provider_engine import (
    ServiceProviderEngine,
)
from wirio._service_lookup._typed_type import TypedType
from wirio.abstractions.base_service_provider import BaseServiceProvider
from wirio.abstractions.service_provider_is_keyed_service import (
    ServiceProviderIsKeyedService,
)
from wirio.abstractions.service_provider_is_service import (
    ServiceProviderIsService,
)
from wirio.abstractions.service_scope import (
    AbstractAsyncContextManager,
    ServiceScope,
)
from wirio.abstractions.service_scope_factory import (
    ServiceScopeFactory,
)
from wirio.exceptions import (
    ObjectDisposedError,
    ServiceProviderNotFullyInitializedError,
)
from wirio.service_provider_engine_scope import (
    ServiceProviderEngineScope,
)

if TYPE_CHECKING:
    from wirio.service_descriptor import ServiceDescriptor


@dataclass(frozen=True)
class _ServiceAccessor:
    call_site: ServiceCallSite | None
    realized_service: Callable[[ServiceProviderEngineScope], Awaitable[object | None]]


@final
class ServiceProvider(
    BaseServiceProvider, AbstractAsyncContextManager["ServiceProvider"]
):
    """Provider that resolves services."""

    _descriptors: Final[list["ServiceDescriptor"]]
    _root: Final[ServiceProviderEngineScope]
    _engine: Final[ServiceProviderEngine]
    _service_accessors: Final[
        AsyncConcurrentDictionary[ServiceIdentifier, _ServiceAccessor]
    ]
    _is_disposed: bool
    _call_site_factory: Final[CallSiteFactory]
    _is_fully_initialized: bool

    def __init__(self, descriptors: list["ServiceDescriptor"]) -> None:
        self._descriptors = descriptors
        self._root = ServiceProviderEngineScope(
            service_provider=self, is_root_scope=True
        )
        self._engine = self._get_engine()
        self._service_accessors = AsyncConcurrentDictionary()
        self._is_disposed = False
        self._call_site_factory = CallSiteFactory(descriptors)
        self._is_fully_initialized = False

    @property
    def root(self) -> ServiceProviderEngineScope:
        return self._root

    @property
    def is_disposed(self) -> bool:
        return self._is_disposed

    @property
    def is_fully_initialized(self) -> bool:
        """Indicate whether the provider is fully initialized (useful for Jupyter notebooks, which don't work well with context managers)."""
        return self._is_fully_initialized

    @override
    async def get_service_object(self, service_type: TypedType) -> object | None:
        if self._is_disposed:
            raise ObjectDisposedError

        await self._fully_initialize_if_not_fully_initialized()
        return await self.get_service_from_service_identifier(
            service_identifier=ServiceIdentifier.from_service_type(service_type),
            service_provider_engine_scope=self._root,
        )

    @override
    async def get_keyed_service_object(
        self, service_key: object | None, service_type: TypedType
    ) -> object | None:
        if self._is_disposed:
            raise ObjectDisposedError

        await self._fully_initialize_if_not_fully_initialized()
        return await self.get_service_from_service_identifier(
            service_identifier=ServiceIdentifier.from_service_type(
                service_type=service_type, service_key=service_key
            ),
            service_provider_engine_scope=self._root,
        )

    def create_scope(self) -> ServiceScope:
        """Create a new :class:`ServiceScope` that can be used to resolve scoped services."""
        if self._is_disposed:
            raise ObjectDisposedError

        self._ensure_is_fully_initialized(self.create_scope.__name__)
        return ServiceProviderEngineScope(service_provider=self, is_root_scope=False)

    async def aclose(self) -> None:
        """Dispose the service provider and release all resources."""
        await self.__aexit__(None, None, None)

    async def get_service_from_service_identifier(
        self,
        service_identifier: ServiceIdentifier,
        service_provider_engine_scope: ServiceProviderEngineScope,
    ) -> object | None:
        override_call_site = self._call_site_factory.get_overridden_call_site(
            service_identifier
        )

        if override_call_site is not None:
            realized_override = self._engine.realize_service(override_call_site)
            return await realized_override(service_provider_engine_scope)

        service_accessor = await self._service_accessors.get_or_add(
            key=service_identifier, value_factory=self._create_service_accessor
        )
        return await service_accessor.realized_service(service_provider_engine_scope)

    @contextmanager
    def override_service(
        self, service_type: type, implementation_instance: object | None
    ) -> Generator[None]:
        """Override a service registration within the context manager scope.

        It can be used to temporarily replace a service for testing specific scenarios. Don't use it in production.
        """
        self._ensure_is_fully_initialized(self.override_service.__name__)
        service_identifier = ServiceIdentifier.from_service_type(
            TypedType.from_type(service_type)
        )

        with self._call_site_factory.override_service(
            service_identifier=service_identifier,
            implementation_instance=implementation_instance,
        ):
            yield

    @contextmanager
    def override_keyed_service(
        self,
        service_key: object | None,
        service_type: type,
        implementation_instance: object | None,
    ) -> Generator[None]:
        """Override a keyed service registration within the context manager scope.

        It can be used to temporarily replace a service for testing specific scenarios. Don't use it in production.
        """
        self._ensure_is_fully_initialized(self.override_keyed_service.__name__)
        service_identifier = ServiceIdentifier.from_service_type(
            service_type=TypedType.from_type(service_type),
            service_key=service_key,
        )

        with self._call_site_factory.override_service(
            service_identifier=service_identifier,
            implementation_instance=implementation_instance,
        ):
            yield

    async def _create_service_accessor(
        self, service_identifier: ServiceIdentifier
    ) -> _ServiceAccessor:
        def realized_service_returning_none(
            _: ServiceProviderEngineScope,
        ) -> Awaitable[object | None]:
            future = asyncio.Future[None]()
            future.set_result(None)
            return future

        call_site = await self._call_site_factory.get_call_site(
            service_identifier, CallSiteChain()
        )

        if call_site is not None:
            realized_service = self._engine.realize_service(call_site)
            return _ServiceAccessor(
                call_site=call_site, realized_service=realized_service
            )

        return _ServiceAccessor(
            call_site=call_site, realized_service=realized_service_returning_none
        )

    def _get_engine(self) -> ServiceProviderEngine:
        return RuntimeServiceProviderEngine.INSTANCE

    async def _add_built_in_services(self) -> None:
        """Add built-in services that aren't part of the list of service descriptors."""
        await self._call_site_factory.add(
            ServiceIdentifier.from_service_type(
                TypedType.from_type(BaseServiceProvider)
            ),
            ServiceProviderCallSite(),
        )
        await self._call_site_factory.add(
            ServiceIdentifier.from_service_type(
                TypedType.from_type(ServiceScopeFactory)
            ),
            ConstantCallSite(
                TypedType.from_type(ServiceScopeFactory),
                self._root,
            ),
        )
        await self._call_site_factory.add(
            ServiceIdentifier.from_service_type(
                TypedType.from_type(ServiceProviderIsService)
            ),
            ConstantCallSite(
                TypedType.from_type(ServiceProviderIsService), self._call_site_factory
            ),
        )
        await self._call_site_factory.add(
            ServiceIdentifier.from_service_type(
                TypedType.from_type(ServiceProviderIsKeyedService)
            ),
            ConstantCallSite(
                TypedType.from_type(ServiceProviderIsKeyedService),
                self._call_site_factory,
            ),
        )

    async def _activate_auto_activated_singletons(self) -> None:
        """Activate all singletons registered with auto_activate=True."""
        for service_descriptor in self._descriptors:
            if not service_descriptor.auto_activate:
                continue

            await self.get_service_from_service_identifier(
                service_identifier=ServiceIdentifier.from_descriptor(
                    service_descriptor
                ),
                service_provider_engine_scope=self._root,
            )

    async def _fully_initialize_if_not_fully_initialized(self) -> None:
        if not self._is_fully_initialized:
            await self.__aenter__()

    def _ensure_is_fully_initialized(self, method_name: str) -> None:
        if not self._is_fully_initialized:
            raise ServiceProviderNotFullyInitializedError(method_name)

    @override
    async def __aenter__(self) -> Self:
        await self._add_built_in_services()
        await self._activate_auto_activated_singletons()
        self._is_fully_initialized = True
        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        self._is_disposed = True
        await self._root.__aexit__(exc_type, exc_val, exc_tb)
